;;; ada-lite-mode.el --- A very light-weight major mode for Ada source. -*- lexical-binding:t -*-
;;
;; Copyright (C) 2022  Free Software Foundation, Inc.
;;
;; Author: Stephen Leake <stephen_leake@stephe-leake.org>
;; Maintainer: Stephen Leake <stephen_leake@stephe-leake.org>
;; Keywords: languages
;;  ada
;; Version: 0.1
;; package-requires: ((emacs "27.0"))
;;
;; This file is part of GNU Emacs.
;;
;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary
;;
;; A very light-weight major mode for Ada source code, suitable for
;; use with an LSP server
;;
;; To use ada-lite-mode for Ada files, put the following in your
;; emacs startup:
;;
;; (require 'ada-lite-mode)
;;
;; For use with an LSP server, configure project-find-functions to
;; return a project whose (project-root) returns the directory
;; containing the main gpr file.  If your project has one main gpr
;; file, and it is in the root directory of your configuration manager
;; workspace, then the default project-try-vc works. For other
;; situations, ada-lite-find-project may be useful.
;;
;; If you are using eglot, and your project requires other gpr files
;; outside the main project directory tree, set the environment
;; variable GPR_PROJECT_PATH, and add:
;;
;; (setq eglot-extend-to-xref t)
;;
;; Then visit an Ada file and enable eglot (only need to enable eglot
;; once):
;; C-x C-f <file>
;; M-x eglot
;;
;; By default this packages assumes the AdaCore ada_language_server;
;; to use with another LSP server for Ada, set ada-lite-server-exec,
;; or change the setting in eglot-server-programs.
;;
;; To install ada_language_server, install gnatstudio from
;; https://www.adacore.com/community, or install the
;; ada_language_server crate in Alire
;; (https://github.com/alire-project/alire/releases).
;;
;; For help with this package, ask on the comp.lang.ada newsgroup, or
;; the ada-mode email list
;; (https://savannah.nongnu.org/projects/ada-mode/).

(defvar ada-lite-mode-syntax-table
  (let ((table (make-syntax-table)))
    ;; (info "(elisp)Syntax Class Table" "*info syntax class table*")
    ;; (info "(elisp) Syntax Flags") for comment start/end
    ;; make-syntax-table sets all alphanumeric to w, etc; so we only
    ;; have to add ada-specific things.

    ;; string brackets. `%' is the obsolete alternative string
    ;; bracket (arm J.2); if we make it syntax class ", it throws
    ;; font-lock and indentation off the track, so we use syntax class
    ;; $.
    (modify-syntax-entry ?%  "$" table)
    (modify-syntax-entry ?\" "\"" table)

    ;; punctuation; operators etc
    (modify-syntax-entry ?#  "." table); based number
    (modify-syntax-entry ?&  "." table)
    (modify-syntax-entry ?*  "." table)
    (modify-syntax-entry ?+  "." table)
    (modify-syntax-entry ?-  ". 12" table); double hyphen as comment start
    (modify-syntax-entry ?.  "." table)
    (modify-syntax-entry ?/  "." table)
    (modify-syntax-entry ?:  "." table)
    (modify-syntax-entry ?<  "." table)
    (modify-syntax-entry ?=  "." table)
    (modify-syntax-entry ?>  "." table)
    (modify-syntax-entry ?@  "." table)
    (modify-syntax-entry ?\' "." table); attribute; see ada-syntax-propertize for character literal
    (modify-syntax-entry ?\; "." table)
    (modify-syntax-entry ?\\ "." table); default is escape; not correct for Ada strings
    (modify-syntax-entry ?\| "." table)

    ;; \f and \n end a comment.
    (modify-syntax-entry ?\f  ">" table)
    (modify-syntax-entry ?\n  ">" table)

    (modify-syntax-entry ?_ "_" table); symbol constituents, not word.

    (modify-syntax-entry ?\( "()" table)
    (modify-syntax-entry ?\) ")(" table)

    ;; skeleton placeholder delimiters; see ada-skel.el. We use generic
    ;; comment delimiter class, not comment starter/comment ender, so
    ;; these can be distinguished from line end.
    (modify-syntax-entry ?{ "!" table)
    (modify-syntax-entry ?} "!" table)

    table
    )
  "Syntax table to be used for editing Ada source code.")

(defconst ada-lite-keywords
  '(;; Ada 83:
    "abort" "abs" "accept" "access" "all" "and" "array" "at" "begin"
    "body" "case" "constant" "declare" "delay" "delta" "digits" "do"
    "else" "elsif" "end" "entry" "exception" "exit" "for" "function"
    "generic" "goto" "if" "in" "is" "limited" "loop" "mod" "new"
    "not" "null" "of" "or" "others" "out" "package" "pragma" "private"
    "procedure" "raise" "range" "record" "rem" "renames" "return"
    "reverse" "select" "separate" "subtype" "task" "terminate" "then"
    "type" "use" "when" "while" "with" "xor"
    ;; Ada 95:
    "abstract" "aliased" "protected" "requeue" "tagged" "until"
    ;; Ada 2001
    "interface" "overriding" "synchronized"
    ;; Ada 2012:
    "some"
    ;; Ada 2022:
    "parallel")
    "List of Ada keywords.")

(defun ada-lite-font-lock-keywords ()
  "Ada mode value for `font-lock-keywords'."
  (list
   (list (concat "\\_<" (regexp-opt ada-lite-keywords t) "\\_>") '(0 font-lock-keyword-face))
   ))

(defun ada-lite-syntax-propertize (start end)
  "For `syntax-propertize-function'.
Assign `syntax-table' properties in region START .. END.
In particular, character constants are set to have string syntax."
  ;; (info "(elisp)Syntax Properties")
  ;;
  ;; called from `syntax-propertize', inside save-excursion with-silent-modifications
  (let ((inhibit-read-only t)
	(inhibit-point-motion-hooks t))
    (goto-char start)
    (save-match-data
      (while (re-search-forward
	      (concat
	       "[^[:alnum:])]\\('\\)[^'\n]\\('\\)"; 1, 2: character literal, not attribute
	       "\\|[^[:alnum:])]\\('''\\)"; 3: character literal '''
	       )
	      end t)
	;; syntax-propertize-extend-region-functions is set to
	;; syntax-propertize-wholelines by default. We assume no
	;; coding standard will permit a character literal at the
	;; start of a line (not preceded by whitespace).
	(cond
	 ((match-beginning 1)
	  (put-text-property
	   (match-beginning 1) (match-end 1) 'syntax-table '(7 . ?'))
	  (put-text-property
	   (match-beginning 2) (match-end 2) 'syntax-table '(7 . ?')))
	 ((match-beginning 3)
	  (put-text-property
	   (match-beginning 3) (1+ (match-beginning 3)) 'syntax-table '(7 . ?'))
	  (put-text-property
	   (1- (match-end 3)) (match-end 3) 'syntax-table '(7 . ?')))
	 )))))

(define-derived-mode ada-lite-mode prog-mode "ada-lite"
  "A major mode for editing Ada code, using eglot."
  :group 'ada

  (set (make-local-variable 'syntax-propertize-function) #'ada-lite-syntax-propertize)
  (syntax-ppss-flush-cache (point-min));; reparse with new function

  (set (make-local-variable 'parse-sexp-ignore-comments) t)
  (set (make-local-variable 'parse-sexp-lookup-properties) t)
  (set 'case-fold-search t); Ada is case insensitive
  (set 'completion-ignore-case t)
  (set (make-local-variable 'comment-start) "--")
  (set (make-local-variable 'comment-end) "")
  (set (make-local-variable 'comment-start-skip) "---*[ \t]*")
  (set (make-local-variable 'comment-multi-line) nil)

  ;; AdaCore standard style (enforced by -gnaty) requires two spaces
  ;; after '--' in comments; this makes it easier to distinguish
  ;; special comments that have something else after '--'
  (set (make-local-variable 'comment-padding) "  ")

  (setq font-lock-defaults
	'(ada-lite-font-lock-keywords ;; keywords
	  nil ;; keywords-only; font-lock set comment, string faces
	  t ;; case-fold
	  ))

  ;; FIXME: eglot-which-function?
  )

(defcustom ada-lite-server-exec nil
  "Location of an Ada language server.
If non-nil, should be an absolute path to an executable for the
server; this allows specifying a development version. See
`ada-lite-find-server' for default behaviour.")

(defun ada-lite-find-server (&optional _interactive)
  ;; in eglot 1.8, eglot--connect calls CONTACT with 1 arg
  ;; in devel eglot, eglot--connect calls CONTACT with no args
  "Find the language server executable.
If `ada-lite-server-exec' is set, uses that. Otherwise defaults to
AdaCore ada_language_server in `exec-path', then in a gnat
installation found in `exec-path'."
  (if ada-lite-server-exec
      (if (file-readable-p ada-lite-server-exec)
	  (list ada-lite-server-exec)
	(user-error "ada-lite-server-exec '%s' not a readable file"
		    ada-lite-server-exec))

    ;; else look for AdaCore ada_language_server
    ;;
    ;; ada_language_server is provided by a GNAT compiler installation, in a
    ;; directory under GNAT/libexec, which is typically not in PATH.
    (let ((gnat (executable-find "gnat"))
	(path exec-path))
    (when gnat
      (setq path (append
		  path
		  (list
		   (expand-file-name
		    (concat (file-name-directory gnat)
			    "../libexec/gnatstudio/als"))))))
    (let ((guess (locate-file "ada_language_server" path exec-suffixes 1)))
      (if guess
	  (list guess)
	(user-error "ada_language_server not found"))))))

;;; projects integration

(defcustom ada-lite-gnat-prj-file nil
  "The GNAT project file for the current ada-lite project.
Must be an absolute file name; the directory is the language
server process directory.")

(defun ada-lite-find-project (_dir)
  "If ada-lite-mode, return ada-lite project.
For `project-find-functions'"
  (and (eq major-mode 'ada-lite-mode)
       ;; We just assume the current Ada_Language_Server handles all
       ;; Ada files. The user must kill and restart the server to
       ;; change projects. FIXME: We sometimes deliberately open a
       ;; file that is not in the same project (for example, to
       ;; compare two versions of a file); what does ALS do then?
       'ada-lite-project))

(cl-defmethod project-root ((_prj (eql ada-lite-project)))
  (unless ada-lite-gnat-prj-file
    ;; ada_language_server must find a .gpr file on startup. Prompting
    ;; for it here doesn't work, so throw an error.
    (user-error "set `ada-lite-gnat-prj-file'."))
  (file-name-directory ada-lite-gnat-prj-file))

;;;;; Global initializations

;;;###autoload
(add-to-list 'auto-mode-alist '("\\.ad[abs]\\'" . ada-lite-mode))

(if (featurep 'eglot)
    (add-to-list 'eglot-server-programs (cons 'ada-lite-mode #'ada-lite-find-server))
  (eval-after-load "eglot" '(add-to-list 'eglot-server-programs (cons 'ada-lite-mode #'ada-lite-find-server))))

(provide 'ada-lite-mode)
;;; ada-lite-mode.el ends here
